from PySide6.QtCore import Qt, QTimer

from infinote.config import Config


def parse_key_event_into_text(event):
    special_keys = {
        32: "Space",
        16777219: "BS",
        16777216: "Esc",
        16777220: "CR",
        16777217: "Tab",
        16777232: "Home",
        16777233: "End",
        16777223: "Del",
        16777237: "Down",
        16777235: "Up",
        16777234: "Left",
        16777236: "Right",
        16777239: "PageDown",
        16777238: "PageUp",
    }
    key = event.key()
    if key in special_keys:
        text = special_keys[key]
    elif key < 0x110000:
        text = chr(key).lower()
    else:
        return

    mods = event.modifiers()
    # if ctrl
    if mods & Qt.ControlModifier:
        text = "C-" + text
    if mods & Qt.ShiftModifier:
        text = "S-" + text
    if mods & Qt.AltModifier:
        text = "A-" + text
    if mods & Qt.MetaModifier:
        text = "M-" + text
    # if mods & Qt.KeypadModifier:
    #     text = "W-" + text

    if mods or (key in special_keys):
        text = "<" + text + ">"

    return text


class KeyHandler:
    def __init__(self, nvim, view):
        self.nvim = nvim
        self.view = view

        self.command = ""
        self.command_mode = False
        self.search_mode = False
        self.backward_search_mode = False
        self._leader_key_last_pressed = False

    def _absorb_key_into_command_line(self, text, event):
        if text == "<Esc>":
            self.command_mode = False
            self.search_mode = False
            self.backward_search_mode = False
            self.command = ""
        elif text == "<BS>":
            self.command = self.command[:-1]
        elif text == "<CR>":
            if self.command_mode:
                self.nvim.input(f":{self.command}<CR>")
            elif self.search_mode:
                self.nvim.input(f"/{self.command}<CR>")
            elif self.backward_search_mode:
                self.nvim.input(f"?{self.command}<CR>")
            self.command_mode = False
            self.search_mode = False
            self.backward_search_mode = False
            self.command = ""
        elif event.text():
            self.command += event.text()

    def handle_key_event(self, event):
        text = parse_key_event_into_text(event)
        if text is None:
            return

        if text in Config.keys:
            # custom command pressed
            self.handle_custom_command(text)
            return

        mode = self.nvim.api.get_mode()["mode"]
        # assert mode != "c", "there should be no way to get into command mode"
        if mode != "n":
            # send that key
            self.nvim.input(text)
            return

        # if we're here, we're in normal mode

        # handle custom commands
        if self._leader_key_last_pressed:
            self._leader_key_last_pressed = False
            self.handle_custom_command(Config.leader_key + text)
            return

        # monitor command and search input
        if self.command_mode or self.search_mode or self.backward_search_mode:
            # eat the keypress into self.command
            self._absorb_key_into_command_line(text, event)
            return

        match text:
            case Config.leader_key:
                self._leader_key_last_pressed = True

            # handle entering command line
            case "<S-:>" | ":":
                self.command_mode = True
            case "/":
                self.search_mode = True
            case "<S-/>" | "<S-?>" | "?":
                self.backward_search_mode = True

            case _:
                # send that key
                self.nvim.input(text)

    def get_command_line(self):
        if self.command_mode:
            return ":" + self.command
        elif self.search_mode:
            return "/" + self.command
        elif self.backward_search_mode:
            return "?" + self.command
        else:
            return ""

    def handle_custom_command(self, key_combo):
        if key_combo not in Config.keys:
            return
        command = Config.keys[key_combo]

        buf_handler = self.view.buf_handler
        view = self.view

        match command:
            case "hop":
                cmd = ":lua require('leap').leap { target_windows = vim.api.nvim_list_wins() }"
                self.nvim.input(f":{cmd}<CR>")
            case "bookmark jump":
                cmd = (
                    '<Home>"fyt|f|<Right>"lyiw:buffer<Space><C-r>f<Enter>:<C-r>l<Enter>'
                )
                self.nvim.input(cmd)
                view.zoom_on_text(buf_handler.get_current_text())
            case "center on current text":
                current_text = buf_handler.get_current_text()
                view.global_scale = view.get_scale_centered_on_text(current_text)
            case "maximize on current text":
                current_text = buf_handler.get_current_text()
                view.global_scale = view.get_scale_maximized_on_text(current_text)
            case "create child down":
                buf_handler.create_child("down")
            case "create child right":
                buf_handler.create_child("right")
            case "move down":
                view.jump_to_neighbor("down")
            case "move up":
                view.jump_to_neighbor("up")
            case "move left":
                view.jump_to_neighbor("left")
            case "move right":
                view.jump_to_neighbor("right")
            case "zoom up":
                if view.timer is None:
                    view.timer = QTimer()
                    view.timer.timeout.connect(lambda: view.zoom(-1))
                    view.timer.start(1000 / Config.FPS)
            case "zoom down":
                if view.timer is None:
                    view.timer = QTimer()
                    view.timer.timeout.connect(lambda: view.zoom(1))
                    view.timer.start(1000 / Config.FPS)
            case "grow box":
                if view.timer is None:
                    view.timer = QTimer()
                    view.timer.timeout.connect(lambda: view.resize(1))
                    view.timer.start(1000 / Config.FPS)
            case "shrink box":
                if view.timer is None:
                    view.timer = QTimer()
                    view.timer.timeout.connect(lambda: view.resize(-1))
                    view.timer.start(1000 / Config.FPS)
            case "jump back":
                buf_handler.jump_back()
                view.zoom_on_text(buf_handler.get_current_text())
            case "jump forward":
                buf_handler.jump_forward()
                view.zoom_on_text(buf_handler.get_current_text())
            case "catch child down":
                buf_handler.catch_child = "down"
                view.msg("catching child down")
            case "catch child right":
                buf_handler.catch_child = "right"
                view.msg("catching child right")



import re

from PySide6.QtGui import QColor, QFont


# (note: the order of modifiers must be M-, A-, S-, C-)
qwerty_keys = {
    # move to neighbors
    "<A-j>": "move down",
    "<A-k>": "move up",
    "<A-h>": "move left",
    "<A-l>": "move right",
    # create a child of the current text box, down of it
    "<M-A-j>": "create child down",
    # create a child of the current text box, right of it
    "<M-A-l>": "create child right",
    # catch a child and insert it down
    "<A-S-j>": "catch child down",
    # catch a child and insert it right
    "<A-S-l>": "catch child right",
    # zooming
    "<A-y>": "zoom down",
    "<A-o>": "zoom up",
    # resizing box
    "<A-u>": "grow box",
    "<A-i>": "shrink box",
}
colemak_keys = {
    # move to neighbors
    "<A-n>": "move down",
    "<A-e>": "move up",
    "<A-m>": "move left",
    "<A-i>": "move right",
    # create a child of the current text box, down of it
    "<M-A-n>": "create child down",
    # create a child of the current text box, right of it
    "<M-A-i>": "create child right",
    # catch a child and insert it down
    "<A-S-n>": "catch child down",
    # catch a child and insert it right
    "<A-S-i>": "catch child right",
    # zooming
    "<A-j>": "zoom down",
    "<A-y>": "zoom up",
    # resizing box
    "<A-l>": "grow box",
    "<A-u>": "shrink box",
}


class Config:
    autoshrink = True
    text_width = 400
    text_max_height = text_width * 1.618
    initial_position = (500, 40)
    text_gap = 6
    starting_box_scale = 0.75

    # closer to 1 is slower (must be larger than 1)
    scroll_speed = 1.0005
    # invert scroll direction
    scroll_invert = False
    # speed of zooming left/right with keys (must be larger than 1)
    key_zoom_speed = 3
    # whether to allow resizing text boxes with mouse wheel
    scroll_can_resize_text = False
    # when jumping to neighbor, if no text is connected in chosen direction,
    # jump to closest disconnected text in that direction
    allow_disconnected_jumps = True

    # whether to change zoom level on jumps to a neighbor text
    track_jumps_on_neighbor_moves = False

    # https://blog.depositphotos.com/15-cyberpunk-color-palettes-for-dystopian-designs.html
    background_color = "#000000"
    border_brightness = "15%"
    text_brightness = "80%"
    selection_brightness = "23%"
    non_persistent_hue = 341
    sign_color = QColor.fromHsl(289, 100, 38)

    leader_key = ","
    # supported single key codes, and single key codes preceded with leader key
    # (note: the order of modifiers must be M-, A-, S-, C-)
    keys = {
        # hop to any text using leap plugin
        ",h": "hop",
        # when in bookmarks window, jump to location of bookmark under cursor
        ",b": "bookmark jump",
        # center view on the current text box
        ",c": "center on current text",
        # zoom in as much as possible, while keeping the current text box in view
        ",m": "maximize on current text",
        # custom C-o and C-i, because normal ones create unwanted buffers
        "<C-o>": "jump back",
        "<C-i>": "jump forward",
    }
    keys.update(qwerty_keys)

    # relevant for zooming and resizing with keys
    FPS = 180

    input_on_creation = "- "

    # font sizes for each indent level
    # font_sizes = [16] * 4 + [14] * 4 + [11] * 4
    # font_sizes = [15] * 4 + [14] * 4 + [11] * 4
    # font_sizes = [14] * 4 + [11] * 4 + [8] * 4
    font_sizes = [14] * 4 + [11] * 4 + [11] * 4
    # font_sizes = [11] * 4 + [8] * 4 + [6] * 4
    # some font sizes cause indent problems:
    # note that the problems also depent on the intended indent of that font
    # the combinations above are one of the very few that work well, so it's
    # recommended to just choose one of those
    # good values for first indent lvl: 16, 15, 14, 11
    # for the second indent level: 14, 11, 8, 6
    # for the third indent level: 14, 11, 8, 6, 5

    # when centering or maximizing on a text, this defines min gap left to win border
    min_gap_win_edge = 0.02

    ########################
    # don't tweak those - those are automatic calculations
    _initial_distance = (initial_position[0] ** 2 + initial_position[1] ** 2) ** 0.5

    fonts = [QFont("monospace", fs) for fs in font_sizes]



from pathlib import Path

from PySide6.QtCore import QPointF, Qt
from PySide6.QtGui import (
    QColor,
    QFontMetrics,
    QTextBlockFormat,
    QTextCharFormat,
    QTextCursor,
)
from PySide6.QtWidgets import QGraphicsProxyWidget, QTextBrowser, QTextEdit

from infinote.config import Config


def is_buf_empty(buf):
    if len(buf) > 1:
        return False
    only_line = buf[0]
    if only_line.strip() == "":
        return True
    return False


# class NonSelectableTextEdit(QTextBrowser):
class NonSelectableTextEdit(QTextEdit):
    def mousePressEvent(self, event):
        # Skip the mouse press event to prevent it from being handled by QTextBrowser
        event.ignore()

    def mouseMoveEvent(self, event):
        # Skip the mouse move event to prevent it from being handled by QTextBrowser
        event.ignore()

    def mouseReleaseEvent(self, event):
        # Skip the mouse release event to prevent it from being handled by QTextBrowser
        event.ignore()

    def keyPressEvent(self, event):
        event.ignore()


class DraggableText(QGraphicsProxyWidget):
    def __init__(self, nvim, buffer_handle, filename, view, plane_pos, manual_scale):
        super().__init__()

        self.autoshrink = Config.autoshrink

        # note that num doesn't need to be the same as buffer_handle.number
        self.buffer = buffer_handle
        self.filename = filename
        self.view = view
        self.plane_pos = plane_pos
        self.manual_scale = manual_scale
        self.setScale(manual_scale)
        self.child_down = None
        self.child_right = None
        self.parent = None
        self.pin_pos = None
        self.folds = []
        self.sign_lines = []
        self._height = 0
        self.cursor_pos = 0

        self.text_box = NonSelectableTextEdit()
        self.text_box.setFixedWidth(Config.text_width)

        # get folds for potential future fold drawing
        assert self.buffer == nvim.current.buffer
        self.folds = nvim.eval("GetAllFolds()")

        signs = nvim.eval("sign_getplaced()")
        self._set_sign_lines(signs)

        # optionally, send some input on creation
        if is_buf_empty(self.buffer) and Config.input_on_creation:
            nvim.command("startinsert")
            nvim.input(Config.input_on_creation)

        if filename is None:
            # it's non-persistent buffer, so mark its border yellow
            hue = Config.non_persistent_hue
        else:
            savedir = Path(filename).parent
            hue = self.view.buf_handler.savedir_hues[savedir]
        style = f"""
            QTextEdit {{
                background-color: {Config.background_color};
                border: 1px solid hsl({hue}, 96%, {Config.border_brightness});
                color: hsl({hue}, 96%, {Config.text_brightness});
            }}
            QScrollBar:vertical {{
                width: 15px;
                background: {Config.background_color};
            }}
            QScrollBar::handle:vertical {{
                background-color: hsl({hue}, 96%, {Config.border_brightness});
            }}
        """
        self.text_box.setStyleSheet(style)

        self.text_color = QColor()
        self.text_color.setHsl(hue, 96, int(Config.text_brightness[:-1]))
        self.selection_color = QColor()
        self.selection_color.setHsl(hue, 96, int(Config.selection_brightness[:-1]))

        doc = self.text_box.document()
        doc.setIndentWidth(1)

        self.setWidget(self.text_box)

    # position related functions:

    def mouseMoveEvent(self, event):
        # drag around
        mouse_end = QPointF(event.screenPos() / self.view.global_scale)
        displacement = self.get_plane_scale() * self.pin_pos
        self.plane_pos = mouse_end - displacement

        self.reposition()
        self.view.dummy.setFocus()

        self.detach_parent()

    def get_plane_scale(self):
        if self.autoshrink:
            # euclidean magniture of plane_pos
            distance = (self.plane_pos.x() ** 2 + self.plane_pos.y() ** 2) ** 0.5
            distance_scale = distance / Config._initial_distance
            return self.manual_scale * distance_scale
        else:
            return self.manual_scale

    def reposition(self):
        global_scale = self.view.global_scale
        self.setScale(self.get_plane_scale() * global_scale)
        self.setPos(self.plane_pos * global_scale)

        # set height
        # for some reason it needs to be done twice, to prevent a glitch
        # only the smaller of those two heights is valid
        height = self._calculate_height()
        self.text_box.setFixedHeight(height)
        height = min(self._calculate_height(), height)
        self.text_box.setFixedHeight(height)
        self._height = height

        self.place_down_children()
        self.place_right_children()

    def _yx_to_pos(self, y, x):
        # get the one number char position
        doc = self.text_box.document()
        return doc.findBlockByLineNumber(y - 1).position() + x

    def detach_parent(self):
        # detach from parent
        if self.parent is not None:
            if self.parent.child_down == self:
                self.parent.child_down = None
            elif self.parent.child_right == self:
                self.parent.child_right = None
            self.parent = None

    def detach_children(self):
        if self.child_down is not None:
            self.child_down.parent = None
            self.child_down = None
        if self.child_right is not None:
            self.child_right.parent = None
            self.child_right = None

    def place_down_children(self):
        if self.child_down is not None:
            height = self.get_plane_height()
            gap = Config.text_gap * self.get_plane_scale() / self.manual_scale
            self.child_down.plane_pos = self.plane_pos + QPointF(0, height + gap)
            self.child_down.reposition()

    def place_right_children(self):
        if self.child_right is not None:
            width = self.get_plane_width()
            gap = Config.text_gap * self.get_plane_scale() / self.manual_scale
            self.child_right.plane_pos = self.plane_pos + QPointF(width + gap, 0)
            self.child_right.reposition()

    def _calculate_height(self):
        height = self.text_box.document().size().height() + 2
        height = min(height, Config.text_max_height)
        return height

    def get_plane_width(self):
        return self.get_plane_scale() * Config.text_width

    def get_plane_height(self):
        # this needs to be called only after this node's reposition
        return self.get_plane_scale() * self._height

    def get_center(self):
        # note: it's in screen coords, not plane coords
        return self.mapToScene(self.rect().center())

    # text related functions:

    def save(self, nvim):
        if self.filename is None:
            # this buffer was not created by this program, so don't save it
            return

        # take the actual filename from the buffer
        buf_filename = self.buffer.name
        buf_filename = Path(buf_filename).resolve().as_posix()
        # make sure the actual filename is the same as given one
        assert buf_filename == self.filename, (buf_filename, self.filename)

        # set this buffer as current
        nvim.api.set_current_buf(self.buffer.number)
        # save it
        nvim.command("w")

    def highlight(self, color, start, end, invert=False):
        if not isinstance(color, QColor):
            color = QColor(color)
        color_format = QTextCharFormat()
        color_format.setBackground(color)
        if invert:
            color_format.setForeground(QColor(Config.background_color))

        start_pos = self._yx_to_pos(start[0], start[1] - 1)
        end_pos = self._yx_to_pos(end[0], end[1])

        cursor = self.text_box.textCursor()
        cursor.setPosition(start_pos, QTextCursor.MoveAnchor)
        cursor.setPosition(end_pos, QTextCursor.KeepAnchor)
        cursor.mergeCharFormat(color_format)

    def _get_blocks(self):
        doc = self.text_box.document()
        block = doc.begin()
        while block.isValid():
            yield block
            block = block.next()

    def _format_displayed_lines(self):
        # set the fancy formatting, with nice indents and decreasing font sizes
        cursor = self.text_box.textCursor()
        block_format = QTextBlockFormat()
        for block in self._get_blocks():
            line = block.text()
            cursor.setPosition(block.position(), QTextCursor.MoveAnchor)
            cursor.setPosition(block.position() + len(line), QTextCursor.KeepAnchor)

            real_indent = len(line) - len(line.lstrip())
            if line == " ":
                # indent was added artificially, (not strictly true, but it's ok)
                real_indent = 0

            if real_indent < len(Config.font_sizes):
                font = Config.fonts[real_indent]
            else:
                font = Config.fonts[-1]

            font_format = QTextCharFormat()
            font_format.setFont(font)
            cursor.setCharFormat(font_format)

            indent_width = QFontMetrics(font).horizontalAdvance(" " * (2 + real_indent))
            block_format.setIndent(indent_width)
            block_format.setTextIndent(-indent_width)
            cursor.setBlockFormat(block_format)

    def update_text(self, lines, extmarks):
        # maybeTODO optimization:
        # only during leaps all of the texts will be redrawn, so that can
        # become really slow - otherwise it's ok
        #  then I could go back to additionally checking if extmark changed

        # add space to empty lines so that cursor can be displayed there
        for i, line in enumerate(lines):
            if line == "":
                lines[i] = " "

        # set marks text (mainly for the leap plugin)
        mark_positions = []
        for _, y, x, details in extmarks:
            virt_text = details["virt_text"]
            assert len(virt_text) == 1, virt_text
            char, type_ = virt_text[0]
            if type_ == "Cursor":
                continue
            # maybeTODO later relax this?
            assert type_ in ["LeapLabelPrimary", "LeapLabelSecondary"], extmarks
            # put that char into text
            lines[y] = lines[y][:x] + char + lines[y][x + 1 :]
            mark_positions.append((y, x))

        # set new text
        new_text = "\n".join(lines)
        self.text_box.setText(new_text)

        self._format_displayed_lines()

        # highlight the chars
        for y, x in mark_positions:
            self.highlight("brown", (y + 1, x + 1), (y + 1, x + 1))

        # clear cursor
        cursor = self.text_box.textCursor()
        cursor.setPosition(0)
        self.text_box.setTextCursor(cursor)

    def update_current_text(self, mode_info, cur_buf_info, lines):
        # this function if called only if this node's buffer is the current buffer
        mode = mode_info["mode"]

        # get folds for potential future fold drawing
        self.folds = cur_buf_info["folds"]

        # get highlight bookmarks for potential future drawing
        self._set_sign_lines(cur_buf_info["bookmark_info"])

        # set selection
        if mode == "v" or mode == "V" or mode == "\x16":
            s = cur_buf_info["selection_start"][1:3]
            e = cur_buf_info["selection_end"][1:3]
            # if end before start, swap
            if s[0] > e[0] or (s[0] == e[0] and s[1] > e[1]):
                s, e = e, s
        if mode == "v":
            self.highlight(self.selection_color, s, e)
        elif mode == "V":
            # extend selection to full line
            s[1] = 1
            e[1] = len(lines[e[0] - 1])
            self.highlight(self.selection_color, s, e)
        elif mode == "\x16":
            # visual block selection
            x_start = min(s[1], e[1])
            x_end = max(s[1], e[1])
            for y in range(s[0], e[0] + 1):
                self.highlight(self.selection_color, (y, x_start), (y, x_end))

    def draw_cursor(self, mode_info, cur_buf_info):
        mode = mode_info["mode"]
        # set cursor
        curs_y, curs_x = cur_buf_info["cursor_position"]
        pos = self._yx_to_pos(curs_y, curs_x)
        if mode == "n":
            _yx_pos = (curs_y, curs_x + 1)
            self.highlight(self.text_color, _yx_pos, _yx_pos, invert=True)
        elif mode == "i":
            # get focus so that cursor is displayed
            self.text_box.setFocus()
        self.cursor_pos = pos

    def hide_folds(self):
        # set folds
        head_lines = set()
        hidden_lines = set()
        for fold in self.folds:
            head_lines.add(fold[0])
            current = fold[0] + 1
            while current <= fold[1]:
                hidden_lines.add(current)
                current += 1

        block = self.text_box.document().begin()
        cursor = self.text_box.textCursor()
        line_num = 1
        while block.isValid():
            if line_num in hidden_lines:
                cursor.setPosition(block.position())
                cursor.select(QTextCursor.BlockUnderCursor)
                block = block.next()
                cursor.removeSelectedText()
            elif line_num in head_lines:
                cursor.setPosition(block.position() + block.length() - 1)
                cursor.insertText(" ...")
                block = block.next()
            else:
                block = block.next()
            line_num += 1

    def draw_sign_lines(self, lines):
        for sign_line in self.sign_lines:
            line_width = len(lines[sign_line - 1])
            self.highlight(Config.sign_color, (sign_line, 1), (sign_line, line_width))

    def _set_sign_lines(self, signs):
        if signs != []:
            signs = signs[0]["signs"]
            self.sign_lines = [sign["lnum"] for sign in signs]
        else:
            self.sign_lines = []

    def set_invisible_cursor_pos(self):
        # to prevent weird line glitches, we need to set always the same cursor font
        # and have it as a normal caret, not selection
        cursor = self.text_box.textCursor()
        cursor.setPosition(self.cursor_pos)
        font = Config.fonts[0]
        font_format = QTextCharFormat()
        font_format.setFont(font)
        cursor.setCharFormat(font_format)
        self.text_box.setTextCursor(cursor)



import time

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QPainter
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsRectItem,
    QGraphicsScene,
    QGraphicsView,
    QStatusBar,
)

from infinote.buffer_handling import BufferHandler
from infinote.config import Config
from infinote.key_handler import KeyHandler
from infinote.text_object import DraggableText


class GraphicView(QGraphicsView):
    def __init__(self, nvim, main_subdir, parent=None):
        super().__init__(parent)
        self.nvim = nvim
        self.setRenderHint(QPainter.Antialiasing)
        self.setBackgroundBrush(QColor(Config.background_color))
        self.setScene(QGraphicsScene())
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.global_scale = 1.0
        self.key_handler = KeyHandler(nvim, self)
        self.buf_handler = BufferHandler(nvim, self)
        self.current_folder = main_subdir
        self.timer = None
        self._timer_last_update = None

        # dummy object so that the text boxes can be unfocused
        dummy = QGraphicsRectItem()
        dummy.setFlag(QGraphicsItem.ItemIsFocusable)
        self.scene().addItem(dummy)
        self.dummy = dummy

        # create a status bar
        screen_size = self.screen().size()
        self.status_bar = QStatusBar()
        # place it at the top and display
        self.status_bar.setGeometry(0, 0, screen_size.width() + 20, 20)
        color = Config.background_color
        self.status_bar.setStyleSheet("QStatusBar{background-color: " + color + ";}")
        self._message = []
        self.scene().addWidget(self.status_bar)

    def _render_status_bar(self):
        if not self.nvim.api.get_mode()["blocking"]:
            num_unbound_buffers = self.buf_handler.get_num_unbound_buffers()
            if num_unbound_buffers > 0:
                self.msg(f"{num_unbound_buffers} unbound buffer exists")

        command_line = self.key_handler.get_command_line()
        if command_line:
            self._message = [command_line] + self._message

        msg_string = " | ".join(self._message)
        self.status_bar.showMessage(msg_string)

    # event handling methods
    # note: there's one more possible event: mouseMoveEvent, but it's handled by texts

    def resizeEvent(self, event):
        self.scene().setSceneRect(0, 0, event.size().width(), event.size().height())
        super().resizeEvent(event)

    def mousePressEvent(self, event):
        self._message = []
        item = self.scene().itemAt(event.screenPos(), self.transform())
        if isinstance(item, DraggableText):
            # clicked on text, so make it current
            self.buf_handler.jump_to_buffer(item.buffer.number)
            self.buf_handler.update_all_texts()

            # pin the click position, in case of dragging
            click_pos = event.screenPos() / self.global_scale
            item.pin_pos = (click_pos - item.plane_pos) / item.get_plane_scale()
        else:
            # clicked bg, so create a new text
            item = self.buf_handler.create_text(
                self.current_folder, event.screenPos() / self.global_scale
            )
            self.buf_handler.update_all_texts()

        super().mousePressEvent(event)
        item.setFocus()
        self._render_status_bar()

    def keyPressEvent(self, event):
        self._message = []
        self.key_handler.handle_key_event(event)
        self.buf_handler.update_all_texts()
        self._render_status_bar()

    def wheelEvent(self, event):
        direction = -1 if Config.scroll_invert else 1
        zoom_factor = Config.scroll_speed ** (event.angleDelta().y() * direction)

        item = self.scene().itemAt(event.position(), self.transform())
        if isinstance(item, DraggableText) and Config.scroll_can_resize_text:
            # zoom it
            item.manual_scale *= zoom_factor
            item.reposition()
        else:
            # zoom the whole view
            self.global_scale *= zoom_factor

            # reposition all texts
            for text in self.buf_handler.get_root_texts():
                text.reposition()

    def msg(self, msg):
        self._message.append(msg)

    def jump_to_neighbor(self, direction: str):
        current_text = self.buf_handler.get_current_text()
        match direction:
            case "down":
                new = current_text.child_down
            case "right":
                new = current_text.child_right
            case "up":
                new = current_text.parent
                if new is not None and new.child_down != current_text:
                    # new is not our up parent
                    new = None
            case "left":
                new = current_text.parent
                if new is not None and new.child_right != current_text:
                    # new is not our left parent
                    new = None

        if new is None and Config.allow_disconnected_jumps:
            new = self._get_closest_text(current_text, direction)

        if new is None:
            return

        buf_num = new.buffer.number
        self.buf_handler.jump_to_buffer(buf_num)

        if Config.track_jumps_on_neighbor_moves:
            self.track_jump(current_text, new)

        self.zoom_on_text(new)

    def track_jump(self, old, new):
        # update global scale to track the movement
        old_pos = old.plane_pos
        old_dist = (old_pos.x() ** 2 + old_pos.y() ** 2) ** 0.5
        new_pos = new.plane_pos
        new_dist = (new_pos.x() ** 2 + new_pos.y() ** 2) ** 0.5

        self.global_scale *= old_dist / new_dist

    def zoom_on_text(self, text):
        # keep the text in view
        smallest_scale = self.get_scale_centered_on_text(text)
        biggest_scale = self.get_scale_maximized_on_text(text)
        self.global_scale = min(biggest_scale, self.global_scale)
        self.global_scale = max(smallest_scale, self.global_scale)

    def get_scale_centered_on_text(self, text):
        x = text.plane_pos.x()
        y = text.plane_pos.y()
        window_width = self.screen().size().width()
        window_height = self.screen().size().height()

        center_scale_x = window_width / (x * 2 + text.get_plane_width())
        center_scale_y = window_height / (y * 2 + text.get_plane_height())

        return min(center_scale_x, center_scale_y)

    def get_scale_maximized_on_text(self, text):
        x = text.plane_pos.x()
        y = text.plane_pos.y()
        window_width = self.screen().size().width()
        window_height = self.screen().size().height()

        width_scale = window_width / (x + text.get_plane_width())
        width_scale *= 1 - Config.min_gap_win_edge * 9 / 16
        height_scale = window_height / (y + text.get_plane_height())
        height_scale *= 1 - Config.min_gap_win_edge

        return min(width_scale, height_scale)

    def keyReleaseEvent(self, event):
        if event.isAutoRepeat():
            return

        if self.timer is not None:
            self.timer.stop()
            self.timer = None
            self._timer_last_update = None

    def zoom(self, sign):
        if self._timer_last_update is None:
            # timer just starting
            self._timer_last_update = time.time()
            return
        new_time = time.time()
        time_diff = new_time - self._timer_last_update
        self._timer_last_update = new_time

        self.global_scale *= Config.key_zoom_speed ** (time_diff * sign)

        # reposition all texts
        for text in self.buf_handler.get_root_texts():
            text.reposition()

    def resize(self, sign):
        if self._timer_last_update is None:
            # timer just starting
            self._timer_last_update = time.time()
            return
        new_time = time.time()
        time_diff = new_time - self._timer_last_update
        self._timer_last_update = new_time

        # resize current text box
        text = self.buf_handler.get_current_text()
        text.manual_scale *= Config.key_zoom_speed ** (time_diff * sign)
        text.reposition()

    def _get_closest_text(self, current_text, direction):
        current_center = current_text.get_center()

        # get all texts in that direction
        candidate_text_distances = dict()
        for text in self.buf_handler.get_texts():
            if text == current_text:
                continue
            diff = text.get_center() - current_center
            x, y = diff.x(), diff.y()

            if (
                (direction == "down" and y >= abs(x))
                or (direction == "right" and x >= abs(y))
                or (direction == "up" and y <= -abs(x))
                or (direction == "left" and x <= -abs(y))
            ):
                candidate_text_distances[text] = (x**2 + y**2) ** 0.5

        if len(candidate_text_distances) == 0:
            return None

        # get the closest one
        return min(candidate_text_distances.keys(), key=candidate_text_distances.get)



from collections import defaultdict, deque
from pathlib import Path
from typing import List

from PySide6.QtCore import QPointF

from infinote.config import Config
from infinote.text_object import DraggableText, is_buf_empty


class BufferHandler:
    def __init__(self, nvim, view):
        self.nvim = nvim

        self.view = view
        self.jumplist = None  # must be set by view
        self.buf_num_to_text = {}
        self.forward_jumplist = []
        self.last_file_nums = defaultdict(lambda: 0)
        self.savedir_hues = {}
        self.to_redraw = set()
        self.catch_child = None

    def get_num_unbound_buffers(self):
        return len(self.nvim.buffers) - len(self.buf_num_to_text)


    def open_filename(self, pos, manual_scale, filename=None, buffer=None):
        if isinstance(pos, (tuple, list)):
            pos = QPointF(*pos)

        if buffer is None and filename is not None:
            # no buffer provided, so open the one with the given filename
            num_of_texts = len(self.buf_num_to_text)
            if num_of_texts == 0:
                self.nvim.command(f"edit {filename}")
            elif num_of_texts == len(self.nvim.buffers):
                # create new file
                # this line is the loading time bottleneck - each call is 36ms
                self.nvim.command(f"tabnew {filename}")
            buffer = self.nvim.current.buffer
        elif buffer is not None and filename is None:
            # buffer provided, so open it
            self.nvim.command("tabnew")
            self.nvim.command(f"buffer {buffer.number}")
            # delete the buffer that was created by tabnew
            self.nvim.command("bwipeout! #")
        else:
            raise ValueError("either buffer or filename must be provided")

        text = DraggableText(self.nvim, buffer, filename, self.view, pos, manual_scale)
        self.view.scene().addItem(text)

        self.buf_num_to_text[buffer.number] = text
        return text

    def create_text(self, savedir, pos, manual_scale=Config.starting_box_scale):
        num_of_texts = len(self.buf_num_to_text)
        if num_of_texts == len(self.nvim.buffers) or num_of_texts == 0:
            self.last_file_nums[savedir] += 1
            filename = f"{savedir}/{self.last_file_nums[savedir]}.md"
            return self.open_filename(pos, manual_scale, filename)

        # some buffer was created some other way than calling create_text,
        # so mark it to not be saved
        filename = None

        # get the unused buffer
        for buf in self.nvim.buffers:
            if buf.number not in self.buf_num_to_text:
                return self.open_filename(pos, manual_scale, filename, buf)
        raise RuntimeError("no unused buffer found")

    def jump_to_buffer(self, buf_num):
        # jumping with ":buf <num>" would make some buffers hidden and break leap
        # so we need to jump to the right tab instead
        self.nvim.command(f"call GoToTabWithBuffer({buf_num})")

    def jump_to_file(self, filename):
        # jumping straight to the file would make some buffers hidden and break leap
        # so we need to jump to the right tab instead
        tab_num = None
        for tab in self.nvim.api.list_tabpages():
            wins = self.nvim.api.tabpage_list_wins(tab)
            assert len(wins) == 1, "each tab must have exactly one window"
            candidate_buf_num = self.nvim.api.win_get_buf(wins[0]).number
            candidate_filename = self.nvim.api.buf_get_name(candidate_buf_num)
            candidate_filename = (
                Path(candidate_filename).relative_to(Path.cwd()).as_posix()
            )
            if candidate_filename == filename:
                # found it
                tab_num = tab.number
                break
        assert tab_num is not None, "file not found"
        self.nvim.command(f"tabnext {tab_num}")

    def delete_buf(self, buf):
        text = self.buf_num_to_text.get(buf.number)

        if text.filename is not None:
            # delete the file
            self.nvim.command(f"call delete('{text.filename}')")

        self.nvim.command(f"bwipeout! {buf.number}")
        self.view.scene().removeItem(text)
        self.buf_num_to_text.pop(buf.number)

        # detach
        text.detach_parent()
        text.detach_children()

        # delete from jumplists
        self.jumplist = [x for x in self.jumplist if x != buf.number]
        self.forward_jumplist = [x for x in self.forward_jumplist if x != buf.number]

        del text

    def get_texts(self):
        yield from self.buf_num_to_text.values()

    def get_root_texts(self):
        roots = set()
        for text in self.get_texts():
            while text.parent is not None:
                text = text.parent
            roots.add(text)
        return roots

    def get_current_text(self):
        return self.buf_num_to_text.get(self.nvim.current.buffer.number)

    def create_child(self, side):
        current_text = self.buf_num_to_text.get(self.nvim.current.buffer.number)

        if current_text.filename is None:
            # it's not a persistent buffer, so it shouldn't have children
            self.view.msg("can't create children for non-persistent buffers")
            return

        if side == "right":
            if current_text.child_right is not None:
                self.view.msg("right child already exists")
                return
            child = self.create_text(self.view.current_folder, (0, 0))
            current_text.child_right = child
        elif side == "down":
            if current_text.child_down is not None:
                self.view.msg("down child already exists")
                return
            child = self.create_text(self.view.current_folder, (0, 0))
            current_text.child_down = child
        else:
            raise ValueError("side must be 'right' or 'down'")

        child.parent = current_text
        child.parent.reposition()

        if Config.track_jumps_on_neighbor_moves:
            self.view.track_jump(current_text, child)

        self.view.zoom_on_text(child)

    def jump_back(self):
        if len(self.jumplist) <= 2:
            return
        old = self.jumplist.pop()
        self.forward_jumplist.append(old)
        self.jump_to_buffer(self.jumplist[-1])
        self.to_redraw.add(old)
        old_buf = self.nvim.buffers[old]
        if is_buf_empty(old_buf) or old_buf[:] == [Config.input_on_creation]:
            self.delete_buf(old_buf)

    def jump_forward(self):
        if len(self.forward_jumplist) == 0:
            return
        old = self.get_current_text().buffer.number
        self.to_redraw.add(old)
        new = self.forward_jumplist.pop()
        self.jumplist.append(new)
        self.jump_to_buffer(new)
        old_buf = self.nvim.buffers[old]
        if is_buf_empty(old_buf) or old_buf[:] == [Config.input_on_creation]:
            self.delete_buf(old_buf)

    def reattach_text(self, parent_text, child_text):
        child_text.detach_parent()

        # we have to check we're not creating a cycle
        parents_root = parent_text
        while parents_root.parent is not None:
            parents_root = parents_root.parent
        childs_root = child_text
        while childs_root.parent is not None:
            childs_root = childs_root.parent
        if childs_root == parents_root:
            self.view.msg("can't reattach - we would create a cycle")
            return

        if self.catch_child == "down":
            parent_text.child_down = child_text
            child_text.parent = parent_text
        elif self.catch_child == "right":
            parent_text.child_right = child_text
            child_text.parent = parent_text

    def _sanitize_buffers(self):
        current_buffer, wins = self.nvim.api.call_atomic(
            [
                ["nvim_get_current_buf", []],
                ["nvim_tabpage_list_wins", [self.nvim.current.tabpage]],
            ]
        )[0]

        # delete last buf if it's empty and unfocused
        if self.jumplist[-1] != current_buffer.number:
            _last_buf = self.nvim.buffers[self.jumplist[-1]]
            if is_buf_empty(_last_buf) or _last_buf[:] == [Config.input_on_creation]:
                self.delete_buf(_last_buf)
            current_buffer = self.nvim.current.buffer

        # make sure current tab has the current buffer
        # get the num of wins in this tab
        if len(wins) != 1:
            bufs_in_tab = {self.nvim.api.win_get_buf(win): win for win in wins}
            unbound_bufs = [
                buf for buf in bufs_in_tab if buf.number not in self.buf_num_to_text
            ]
            for unb_buf in unbound_bufs:
                # delete its window
                win = bufs_in_tab[unb_buf]
                self.nvim.api.win_close(win, True)
            current_buffer = self.nvim.current.buffer

        # if hidden buffer focused, focus on the last chosen text
        if current_buffer.number not in self.buf_num_to_text:
            self.jump_to_buffer(self.jumplist[-1])
            current_buffer = self.nvim.current.buffer

        return current_buffer

    def _batched_get_nvim_info(self, to_fetch: List[int]):
        # get all relevalt data in a batched call
        functions = [
            ["nvim_eval", ["GetAllFolds()"]],
            ["nvim_eval", ['getpos("v")']],
            ["nvim_eval", ['getpos(".")']],
            ["nvim_win_get_cursor", [0]],
            ["nvim_eval", ["sign_getplaced()"]],
        ]
        for buf_num in to_fetch:
            functions.append(["nvim_buf_get_lines", [buf_num, 0, -1, False]])

        for buf_num in to_fetch:
            _args = [buf_num, -1, (0, 0), (-1, -1), {"details": True}]
            functions.append(["nvim_buf_get_extmarks", _args])

        # later also get highlight info of bookmarks

        results, errors = self.nvim.api.call_atomic(functions)
        assert errors is None, errors
        results = deque(results)

        cur_buf_info = dict(
            folds=results.popleft(),
            selection_start=results.popleft(),
            selection_end=results.popleft(),
            cursor_position=results.popleft(),
            bookmark_info=results.popleft(),
        )

        all_lines = dict()
        for buf_num in to_fetch:
            all_lines[buf_num] = results.popleft()

        all_extmarks = dict()
        for buf_num in to_fetch:
            all_extmarks[buf_num] = results.popleft()

        return cur_buf_info, all_lines, all_extmarks

    def update_all_texts(self):
        mode_info = self.nvim.api.get_mode()
        if mode_info["blocking"]:
            return

        # unfocus the text boxes - but better would be to always have focus
        self.view.dummy.setFocus()
        self.to_redraw.add(self.jumplist[-1])

        # (note: sanitize_buffers can change the current buffer)
        current_buf = self._sanitize_buffers()

        # catch children
        if self.jumplist[-1] != current_buf.number and self.catch_child is not None:
            parent_text = self.buf_num_to_text[self.jumplist[-1]]
            child_text = self.buf_num_to_text[current_buf.number]
            self.reattach_text(parent_text, child_text)
            self.catch_child = None

        # grow jumplist
        if (
            current_buf.number != self.jumplist[-1]
            and current_buf.number in self.buf_num_to_text
        ):
            self.jumplist.append(current_buf.number)
            self.forward_jumplist = []
            self.jumplist = self.jumplist[-30:]

        self._redraw(mode_info, current_buf)

    def _redraw(self, mode_info, current_buf):
        # choose which ones to redraw
        self.to_redraw.add(current_buf.number)
        all_bufs = self.buf_num_to_text.keys()
        to_redraw = self.to_redraw & set(all_bufs)
        self.to_redraw = set()

        # note that in principle other buffers could have marks
        # but when it comes to leap marks, they are only present iff current buffer
        # has some too
        get_extmarks = [] != self.nvim.api.buf_get_extmarks(
            current_buf.number, -1, (0, 0), (-1, -1), {}
        )
        # potential simplification would be to always get info from all bufs
        # but that may have overhead when there is some huge num of bufs?

        # get batched info from nvim about potentially changed buffers
        cur_buf_info, all_lines, all_extmarks = self._batched_get_nvim_info(
            all_bufs if get_extmarks else list(to_redraw)
        )
        for buf_num, extmarks in all_extmarks.items():
            if extmarks != []:
                to_redraw.add(buf_num)

        ####################################################
        # redraw itself

        # initial redraw
        for buf_num in to_redraw:
            text = self.buf_num_to_text[buf_num]
            lines = all_lines[buf_num]
            extmarks = all_extmarks[buf_num]
            text.update_text(lines, extmarks)

        # draw the things that current buffer has
        current_text = self.buf_num_to_text[current_buf.number]
        lines = all_lines[current_buf.number]
        current_text.update_current_text(mode_info, cur_buf_info, lines)

        # draw sign lines
        for buf_num in to_redraw:
            text = self.buf_num_to_text[buf_num]
            text.draw_sign_lines(all_lines[buf_num])

        # draw cursor in current
        current_text.draw_cursor(mode_info, cur_buf_info)

        # set (invisible) cursor, hide folds
        for buf_num in to_redraw:
            text = self.buf_num_to_text[buf_num]
            # hide folds deletes lines so it needs to be at the end
            text.set_invisible_cursor_pos()
            text.hide_folds()

        # reposition all text boxes
        for text in self.get_root_texts():
            text.reposition()

        ####################################################

        # mark texts with extmarks to be redrawn on next call
        for buf_num, extmarks in all_extmarks.items():
            if extmarks != []:
                self.to_redraw.add(buf_num)



import os
import sys
from pathlib import Path

import pynvim
from PySide6.QtWidgets import QApplication, QMainWindow

from infinote.persistence import load_scene, save_scene
from infinote.view import GraphicView


class MainWindow(QMainWindow):
    def __init__(self, view):
        super().__init__()
        self.view = view
        self.setCentralWidget(self.view)

        # self.resize(1900, 600)
        # self.showMaximized()  # this has small glitches when dragging or zooming
        self.showFullScreen()

        self.show()


def main():
    assert len(sys.argv) == 2, "usage: infinote <savedir>"
    group_dir = Path(sys.argv[1]).resolve()
    workspace_dir = group_dir.parent
    # change working directory to the workspace directory
    # so that nvim can find the bookmark file
    
    workspace_dir.mkdir(parents=True, exist_ok=True)
    os.chdir(workspace_dir)
    print(workspace_dir)

    nvim = pynvim.attach(
        "child", argv=["/usr/bin/env", "nvim", "--embed", "--headless"]
    )
    # text that doesn't fit in window can't be jumped to with Leap (for now)
    nvim.ui_attach(80, 100, True)
    # nvim = pynvim.attach('socket', path='/tmp/nvim')  # there's no speedup to this

    app = QApplication(sys.argv)
    view = GraphicView(nvim, group_dir)
    buf_handler = view.buf_handler
    w = MainWindow(view)

    # make the cursor non-blinking
    app.setCursorFlashTime(0)

    assert len(nvim.buffers) == 1, "we require nvim to start with one buffer"

    load_scene(buf_handler, group_dir)
    view.global_scale = view.get_scale_centered_on_text(buf_handler.get_current_text())
    buf_handler.to_redraw.update(buf_handler.buf_num_to_text.keys())

    buf_handler.jumplist = [None, nvim.current.buffer.number]
    buf_handler.update_all_texts()

    exit_code = app.exec()
    save_scene(buf_handler, nvim, group_dir)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import json
from pathlib import Path

from colormath.color_conversions import convert_color
from colormath.color_objects import HSLColor, LCHabColor
from pynvim import Nvim

from infinote.buffer_handling import BufferHandler
from infinote.config import Config


def _name_to_hue(name: str):
    # choose the hue in a perceptually uniform way
    # choose a num between 60 and 310 degrees, to avoid non-persistent's red
    uniform = (name.__hash__() % 250) + 60  # note: this hash is changing
    random_lch_color = LCHabColor(100, 128, uniform)
    random_HSL_color = convert_color(random_lch_color, HSLColor)
    hue = int(random_HSL_color.hsl_h)
    return hue


def load_scene(buf_handler: BufferHandler, group_dir: Path):
    filename_to_text = {}
    top_dir = group_dir.parent
    top_dir.mkdir(parents=True, exist_ok=True)
    meta = {}

    if not group_dir.exists():
        # save some initial hue for this dir
        hue = _name_to_hue(group_dir.stem)
        meta[group_dir.name] = dict(hue=hue)
        buf_handler.savedir_hues[group_dir] = hue

    meta_path = top_dir / "meta.json"
    if not meta_path.exists():
        print(f"opening a new workspace in {top_dir}")
        # if there is no meta, top_dir should be empty
        assert not any(top_dir.iterdir()), f"workdir_dir not empty: {top_dir}"
        # create the main subdir
        group_dir.mkdir(exist_ok=True)
        # create one text
        buf_handler.create_text(group_dir, Config.initial_position)
        return

    # create the main subdir
    group_dir.mkdir(exist_ok=True)
    meta.update(json.loads(meta_path.read_text()))
    subdirs = [d for d in top_dir.iterdir() if d.is_dir()]
    print(f"subdirs: {subdirs}")
    for subdir in subdirs:
        # load dir color
        assert subdir.name in meta, f"alien folder: {subdir}"
        buf_handler.savedir_hues[subdir] = meta[subdir.name]["hue"]

        # load files into buffers
        files = [f for f in subdir.iterdir() if f.suffix == ".md"]
        for full_filename in files:
            rel_filename = full_filename.relative_to(top_dir).as_posix()
            assert (
                full_filename.stem.isnumeric()
            ), f"names must be integers: {rel_filename}"
            assert rel_filename in meta, f"alien file: {rel_filename}"
            info = meta[rel_filename]

            # create text
            text = buf_handler.open_filename(
                info["plane_pos"], info["manual_scale"], full_filename.as_posix()
            )
            filename_to_text[rel_filename] = text

        # prepare the next file number
        
        max_filenum = max(int(f.stem) for f in files) if files else 0
        buf_handler.last_file_nums[subdir] = max_filenum

    # connect them
    for rel_filename, text in filename_to_text.items():
        info = meta[rel_filename]
        text.child_down = filename_to_text.get(info["child_down"])
        text.child_right = filename_to_text.get(info["child_right"])
        text.parent = filename_to_text.get(info["parent"])
    print(f"loaded {len(filename_to_text)} texts")


def save_scene(buf_handler: BufferHandler, nvim: Nvim, group_dir: Path):
    top_dir = group_dir.parent
    # record text metadata
    meta = {}
    for text in buf_handler.get_texts():
        if text.filename is None:
            # this buffer was not created by this program, so don't save it
            continue
        savedir = Path(text.filename).parent
        rel_filename = Path(text.filename).relative_to(top_dir).as_posix()
        meta[rel_filename] = dict(
            plane_pos=tuple(text.plane_pos.toTuple()),
            manual_scale=text.manual_scale,
            child_down=Path(text.child_down.filename).relative_to(top_dir).as_posix()
            if text.child_down
            else None,
            child_right=Path(text.child_right.filename).relative_to(top_dir).as_posix()
            if text.child_right
            else None,
            parent=Path(text.parent.filename).relative_to(top_dir).as_posix()
            if text.parent
            else None,
        )

    # record other data
    for subdir, hue in buf_handler.savedir_hues.items():
        meta["hue"] = buf_handler.savedir_hues[subdir]
        meta[subdir.name] = dict(hue=hue)

    # save metadata json
    meta_path = top_dir / "meta.json"
    meta_path.write_text(json.dumps(meta, indent=4))

    #########################################
    # save each text
    for text in buf_handler.get_texts():
        text.save(nvim)



from setuptools import find_packages, setup

setup(
    name="infinote-md",
    version="0.2.2",  # Update the version number for new releases
    author="Filip Sondej",
    description="Feel the spatial freedom in your notes.",
    long_description_content_type="text/markdown",
    url="https://github.com/filyp/infinote",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "PySide6",
        "pynvim",
        "colormath",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    package_data={
        "infinote": ["required.vim"],
    },
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "infinote=infinote.main:main",
        ],
    },
)
